/*  gtail
 *  Copyright (C) 1999 Didier Remy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/*
 * $Log: file.c,v $
 * Revision 1.6  1999/12/30 09:00:39  didier
 * Internationalisation
 *
 * Revision 1.5  1999/12/21 18:21:10  didier
 * Mmorisation des PIDs pour la gestion des commandes
 *
 * Dcouplage de la gestion des fichiers et des commandes
 *
 * Gestion de la destruction d'une vue avec la destruction
 *   des informations associes dans notre structure
 *
 * Revision 1.4  1999/11/20 15:42:43  didier
 * Contrle des allocations mmoire (voir les nouvelles options de configuration)
 *
 * Ajout du nom du fichier dans la structure xfileinfo
 *
 * Meilleur vrification de l'existence du fichier  ouvrir
 *
 * Ajout de ma fonction (xt)basename, celle de Linux/Debian n'est pas documente
 *
 * Revision 1.3  1999/11/10 20:34:04  didier
 * Libration de mmoire non alloue en quittant sans avoir ouvert de fichier
 *
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h> /* pour fork() */
#include <signal.h> /* pour signal() */

#include "support.h" /* pour intl */
#include "file.h"
#include "option.h"

/* pas de rfrence  gtk ici, les callback appelant font les
 * cration/destruction li aux oprations sur xfi
 */

xfileinfo *xfi = NULL;

bool_t
init_fi()
{
  int i;
  bool_t res = FALSE;

  if ((xfi = malloc(XFIBLOCKSIZE * sizeof(xfileinfo))) == NULL)
  {
    fprintf(stderr, _("init_fi:can't alloc %dx%d bytes of memory\n"),
	    XFIBLOCKSIZE, sizeof(xfileinfo));
  }
  else
  {
    for (i = 0; i < XFIBLOCKSIZE; i++)
    {
      xfi[i].fd = -1; /* code d'erreur de open */
      xfi[i].pid = -1;
      xfi[i].has_changed = UNUSED;
      xfi[i].w = NULL;
      xfi[i].filename = NULL;
    }
    xfi[XFIBLOCKSIZE - 1].has_changed = LAST;
    res = TRUE;
  }

  return res;
}

#if 0
/* return the number of identical file name 0=none */
int
count_same(const char *s, int lf)
{
  int res = 0, i = 0;
  char *c, *d;

  /* treat LFSELECT and LFLAST the same */
  if (LFSELECT == lf)
    lf = LFLAST;
  while (xfi[i].has_changed != LAST)
  {
    if (LFFULL == lf)
    {
      if (strcmp(xfi[i].filename, s) == 0)
	res++;
    }
    else
    {
      c = strrchr(xfi[i].filename, '/');
      d = strrchr(s, '/');
      if (c == NULL)
	c = xfi[i].filename;
      if (d == NULL)
	d = s;
      if (strcmp(c, d) == 0)
	res++;
    }
    i++;
  }

  return res;
}
#endif

int
add_stream(char *filename, int mode, void *w)
{
  int i, nb, find = 0;

  if (!xfi && !init_fi())
    return -1;

  i = 0;
  while (!find)
  {
    switch (xfi[i].has_changed)
    {
    case LAST: /* le premier lt libre est le dernier allou */
      /* le nombre de bloc deja alloue */
      nb = (i + 1) / XFIBLOCKSIZE + 1; /* division exacte logiquement */
      /*printf ("i:%d nb:%d\n", i, nb);*/
      /*nb++; /* realloue un de plus */
      xfi = realloc(xfi, nb * XFIBLOCKSIZE * sizeof(xfileinfo));
      for (nb = i + 1; nb <= i + XFIBLOCKSIZE; nb++)
      {
	xfi[nb].fd = -1; /* code d'erreur de open */
	xfi[nb].pid = (pid_t)-1;
	xfi[nb].has_changed = UNUSED;
	xfi[nb].w = NULL;
	xfi[nb].filename = NULL;
      }
      /*xfi[i + XFIBLOCKSIZE].has_changed = LAST;*/
      /*printf ("i:%d nb:%d\n", i, nb);*/
      xfi[nb - 1].has_changed = LAST;
      /* on enchaine sur la suite puisqu'on vient de cre de l'espace */
    case UNUSED:
      find = 1;
      switch (mode)
      {
      case /*FILE*/ 0:
	xfi[i].fd = add_file(filename);
	break;
      case /*PIPE*/ 1:
	xfi[i].fd = add_pipe(filename, &xfi[i].pid);
	break;
      default:
	xfi[i].fd = -1;
	fprintf(stderr,
		_("add_file:internal error:not a file nor a command\n"));
	break;
      }
      /* avant cette ligne, on a seulement mis  jour .fd si russit */
      if (xfi[i].fd == -1)
      {
	fprintf(stderr, _("add_file:can't monitor this stream %s\n"),
		filename);
	return -1;
      }
      if ((xfi[i].filename = strdup(filename)) == NULL)
      {
	fprintf(stderr, "add_file:can't allocate %d bytes of mem\n",
		strlen(filename) + 1);
	return -1;
      }
      /* ouf tout va bien on peut finir de changer l'etat de cette entre */
      /* pour les cmd, l'tat DEAD est gr par un signal */
      xfi[i].has_changed = NOT_CHANGED;
      xfi[i].w = w;
#if 0
      xfi[i].last_new_line = 0; /* commence au dbut du buffer */
      xfi[i].idem = count_same(xfi[i].filename,
			       options[which_opt].label_format);
      /* repeat_time et elapse_time sont initialis par le callback associ au
       * bouton "ok" du dialog commande */
#endif
      break;
    default: /* deja utilise */
    /*case IS_DEAD:
    case IS_DEADNCHANGED:
    case CHANGED:
    case NOT_CHANGED:*/
      i++;
    }
  }

  return i;
}

void
remove_file(int ind)
{
  /* libre les ressources avant d'craser par dcalage */
  /* c'est le callback appelant qui dtruit le widget */
  close(xfi[ind].fd);
  free(xfi[ind].filename);
  /* dcale vers la droite pour boucher le trou */
  while (xfi[ind].has_changed != LAST && xfi[ind].has_changed != UNUSED)
  {
    xfi[ind] = xfi[ind + 1];
    ind++;
  }
  /* met le dernier lments dcal xfi[ind]  vide */
  if (xfi[ind].has_changed != LAST) /* en principe pas possible ! */
    xfi[ind].has_changed = UNUSED;
  else
    printf(_("remove_file:internal error:bug last\n"));
  xfi[ind].fd = -1;
  xfi[ind].pid = (pid_t)-1;
  xfi[ind].w = NULL;
  xfi[ind].filename = NULL;
}

void
xfi_free(void)
{
  int i;
  int go = 1;

  if (!xfi) /* on quitte sans avoir rien ouvert */
    return;

  i = 0;
  while (go)
  {
    switch (xfi[i].has_changed)
    {
    case UNUSED:
      break;
    case LAST: /* la fin */
      go = 0;
      /* si le dernier est utilis, il faut aussi le librer et donc pas ne pas
       * faire le break */
      if (xfi[i].fd == -1)
	break;
    case CHANGED:
    case NOT_CHANGED:
    case IS_DEAD:
    case IS_DEADNCHANGED:
      if (xfi[i].pid != -1 && kill(xfi[i].pid, SIGTERM) == -1)
      {
	fprintf(stderr, _("Trying to SIGTERM %s(%d) leads to : "),
		xfi[i].filename, xfi[i].pid);
	perror("xfi_free");
      }
      close(xfi[i].fd);
      free(xfi[i].filename);
    }
    i++;
  }

  free(xfi);
  xfi = NULL;
}

/* renvoie le rpertoire de filename
 * "." si pas de rpertoire NULL
 * si echec */
char *dirname(char *filename)
{
  char *slash, *res;
  int len;

  slash = rindex(filename, '/');
  if (!slash)
    res = strdup(".");
  else
  {
    len = slash - filename;
    res = (char *)malloc(len * sizeof(char));
    if (res)
    {
      strncpy(res, slash, len);
      res[len] = '\0';
    } /* else on va fair return NULL <=> erreur */
  }

  return res;
}

/* dj fournit sous linux mais pas document */
char *xtbasename(const char *filename)
{
  char *slash, *res;
  int len;

  slash = rindex(filename, '/');
  if (!slash)
    res = strdup(filename);
  else
  {
    slash++;
    /* calcul la taille du mot aprs le dernier '/' (+1 avec le 0 terminal)*/
    len = strlen(filename) - (slash - filename) + 1;
    res = (char *)malloc(len * sizeof(char));
    strcpy(res, slash);
  }

  return res;
}

int
add_file(const char * filename)
{
  struct stat s;
  int status;
  int fd = -1;

  /* inspir de support.c:check_file_exists fournit par glade */
  status = stat (filename, &s);
  if (!(status == 0 && S_ISREG (s.st_mode)))
    fprintf(stderr, _("add_file:%s not a file\n"), filename);
  else if ((fd = open(filename, O_RDONLY)) == -1)
    fprintf(stderr, _("add_file:can't open file %s\n"), filename);

  return fd; /* =-1 si un des if echoue */
}

/* une fonction qui fork pour excuter une commande. Elle donne le descripteur
 * de fichier pour y lire son rsultat. si on passe p != NULL, elle y met le
 * pid de la commande */
int
add_pipe(const char * cmd, pid_t * p)
{
  pid_t pid;
  int filedes[2];
  extern void a_child_is_dead(int);

  if (pipe(filedes) == -1)
  {
    fprintf(stderr, _("add_pipe:pipe() failed\n"));
    return -1;
  }

  pid = fork ();

  if (pid > 0)
  {
    /* Processus pere      */
    if (p)
      *p = pid;
    close(filedes[1]);
    signal(SIGCHLD, a_child_is_dead);
    return filedes[0];
  }
  else if (pid == 0)
  {
    /* Processus fils      */
    char * argv [4];

    /* Pris dans ktail 0.4.4 (C) 1999 by Rolf Jakob */
    /* stdin, stdout, et stderr valent 0, 1, and 2 */
    /* le but du jeux ici est de faire correspondre filedes[1] avec stdout */
    /* ferme stdin */
    close(0);
    /* dummy open */
    open("/dev/null", O_RDONLY);
    /* ferme le pipe en lecture */
    close(filedes[0]);
    /* ferme stdout */
    close(1);
    /* duplique le pipe en criture qui doit correspondre avec stdout
     * maintenant */
    dup(filedes[1]);
    /* ferme stdout */
    close(2);
    dup(1);

    argv [0] = "sh";
    argv [1] = "-c";
    argv [2] = strdup(cmd);
    argv [3] = 0;
    execve ("/bin/sh", argv, NULL);
    /* on ne passe ici que si 'sh' ne peut tre lanc ou erreur de execve */
    exit (127);
  }
  else
  {
    /* Traitement d'erreur */
    close(filedes[0]);
    close(filedes[1]);
    fprintf(stderr, _("add_pipe:fork() failed\n"));
    return -1;
  }
}
